;;; hypersrc-read.el -- Functions to read and parse .hs files that
;;; contain HyperSource markup (subset of HTML3)

;;; $Id: hypersrc-read.el,v 0.3 1995/05/17 09:17:51 job Exp job $

(require 'hypersrc-graphics)		; XPM-GIF conversion
(require 'hypersrc-comments)		; correct comment delimiters

(defun hypersrc-HTML-tags ()
  "Return list of HTML tags in current buffer.
HTML tags are strings between < and >.  Each tag is represented as a
list, (BEGINNING END TAG-TEXT).  TAG-TEXT does not include the </>
brackets. BEGINNING points at <, END points behind >. Point is not
moved."
  (save-excursion
    (setq html-tags ())
    (goto-char (point-min))
    (while (re-search-forward "<\\([^>]*\\)>" nil 1) 
	(setq
	 html-tags
	 (append
	  html-tags
	  (list
	   (list
	    (match-beginning 0)
	    (match-end 0)
	    (buffer-substring (match-beginning 1) (match-end 1))))))))
  html-tags)


(defun hypersrc-replace-escapes ()
  "Replace '&lt;', '&gt;', and '&amp;' by '<', '>', and '&', respectively.
Uses current buffer. Point is not moved."
  (save-excursion
    (goto-char (point-min))
    (perform-replace "&lt;" "<" nil nil nil)
    (goto-char (point-min))
    (perform-replace "&gt;" ">" nil nil nil)
    (goto-char (point-min))
    (perform-replace "&amp;" "&" nil nil nil)))


(defun hypersrc-widen-shift-extents ()
  "Move all zero-width hypersrc extents in the buffer forward by one
character, and extend them to width 1 by stretching their right end out by
one further character.
While parsing, margin images and margin comments are inserted one char
before their actual position so they do not move with subsequent text
inserted, and they have a width of 0 which makes them
non-duplicable. These things are fixed with this function after reading
the complete file."
(map-extents
 (function
  (lambda
    (extent maparg)
    (if (and (extent-property extent 'hypersrc)
	     (eq (extent-length extent) 0))
	(set-extent-endpoints
	 extent
	 (+ (extent-start-position extent) 1)
	 (+ (extent-end-position extent) 2)))
    nil))))


(defun hypersrc-initialize-tag-iterators ()
  "Set the 'new tag' iterator variables to nil so storing them later
does not fail."
  (setq new-beg nil new-end nil new-tag nil))


(defun hypersrc-get-new-tag (&optional tag-regexp-to-match)
  "Set the tag iterator variables to the next HTML tag in the list.
Optional argument TAG-REGEXP-TO-MATCH can specify a regular expression
describing the tag that has to be found next. If the next tag in the
list does not match the regexp, parsing is aborted with an error message
saying that the HyperSource file is corrupt."
  ;; Remember old tag first:
  (setq old-beg new-beg old-end new-end old-tag new-tag)
  (setq new-tag-entry (car html-tags))
  ;; Signal error if wrong tag has been found:
  (if (and tag-regexp-to-match
	   (not (string-match tag-regexp-to-match (elt new-tag-entry 2))))
      (error "%s" "HyperSource file corrupt. Parsing failed.")
    (setq new-beg (elt new-tag-entry 0)
	  new-end (elt new-tag-entry 1)
	  new-tag (elt new-tag-entry 2))
    (setq html-tags (cdr html-tags))))


(defun hypersrc-text-between-tags ()
  "Return original buffer contents between end of last end beginning of
next HTML tag."
  (save-excursion
    (set-buffer orig-buf) (buffer-substring old-end new-beg)))


(defun hypersrc-parse-buffer ()
  (let (orig-buf hypersrc-current-major-mode temp-buf html-tags state header-size ext-beg ext-endheader current-annotation comment-text text-beg text-end target-name image-name pixmap anno target new-link target-file-name target-id-string id-string new-target temp-width)
    (setq orig-buf (current-buffer))
    (setq hypersrc-current-major-mode major-mode)
    (setq temp-buf (generate-new-buffer (buffer-name)))
    (set-buffer temp-buf)
    (setq hypersrc-original-left-margin-width (buffer-left-margin-width))
    (set-buffer-left-margin-width hypersrc-left-margin-width)
    (setq use-left-overflow t)
    (setq html-tags (save-excursion (set-buffer orig-buf) (hypersrc-HTML-tags)))
    (hypersrc-initialize-tag-iterators)
    (setq state 'start)
    (while html-tags
      (hypersrc-get-new-tag)		; Set iterator variables to next HTML
										; tag
      (cond
       
       ;; In start state, ignore everything except TABLE tag, up to the
       ;; PRE tag:
       ((eq state 'start)
	(cond
	 ((string-match "^TABLE COLSPEC=\"R" new-tag)
	  (progn
	    (setq hypersrc-left-margin-width 
		  (string-to-int (substring new-tag (match-end 0))))
	    (set-buffer-left-margin-width hypersrc-left-margin-width)))
	 
	 ((string-match "^PRE$" new-tag)
	  (setq state 'pre))))
       
       ;; In pre state, wait for link, target, or /PRE
       ((eq state 'pre)
	(cond
	 ((string= "/PRE" new-tag)	; /PRE -> copy, then check what's next
	  (progn				; (header, comment, image, or /table)!
	    (insert (hypersrc-text-between-tags))
	    (hypersrc-get-new-tag)
	    (cond
	     ((string-match "^H\\([1-6]\\)$" new-tag)
	      (progn
		(setq header-size (string-to-int (substring
						  new-tag (match-beginning 1))))
		(hypersrc-get-new-tag "^/H\\([1-6]\\)$") ; Get "/Hn"
		(setq ext-beg (point))
		(insert (hypersrc-text-between-tags))
		(setq ext-end (point))
		(setq header (make-extent ext-beg ext-end))
		(set-extent-priority header 1)
		(set-extent-property header 'duplicable t)
		(set-extent-property header 'hypersrc 'header)
		(set-extent-property header 'headersize header-size)
		(set-extent-face
		 header
		 (cond
		  ((eq header-size 1)  'hypersrc-header1-face)
		  ((eq header-size 2)  'hypersrc-header2-face)
		  ((eq header-size 3)  'hypersrc-header3-face)
		  ((eq header-size 4)  'hypersrc-header4-face)
		  ((eq header-size 5)  'hypersrc-header5-face)
		  ((eq header-size 6)  'hypersrc-header6-face)))
		(hypersrc-get-new-tag "^PRE$") ; Get "PRE"
		(setq state 'pre)))
	     ((string-match "^TR" new-tag)
	      (progn
		(hypersrc-get-new-tag)
		(cond
		 ((string-match "^TD CLASS=COMMENT" new-tag)
		  (progn
		    (hypersrc-get-new-tag "^SMALL$") ; Skip the "SMALL" tag
		    (hypersrc-get-new-tag "^/SMALL$") ; Get the "/SMALL" tag
		    ;; The annotation has to be shifted back one char
		    ;; temporarily.  Otherwise it sticks to the end of the
		    ;; buffer and moves with all text inserted later.
		    (save-excursion
		      (goto-char (1- (point)))
		      (setq current-annotation
			    (hypersrc-create-new-marginal-comment)))
		    ;; Retrieve comment text, removing comment delimiters:
		    (setq comment-text (hypersrc-text-between-tags))
		    (string-match
		     (regexp-quote
		      (hypersrc-open-comment hypersrc-current-major-mode))
		     comment-text)
		    (setq text-beg (match-end 0))
		    (string-match
		     (regexp-quote
		      (hypersrc-close-comment hypersrc-current-major-mode))
		     comment-text)
		    (setq text-end (match-beginning 0))
		    (setq comment-text
			  (substring comment-text text-beg text-end))
		    (set-annotation-glyph current-annotation comment-text)))
		 ((string-match "^TD CLASS=IMAGE" new-tag)
		  (progn
		    (hypersrc-get-new-tag "^A HREF")
		    (string-match "^A HREF=\"\\([^\"]+\\)\"" new-tag)
		    (setq target-name
			  (substring
			   new-tag (match-beginning 1) (match-end 1)))
		    (setq pixmap (hypersrc-convert-any-to-pixmap target-name))
		    (setq anno (save-excursion
				 (goto-char (1- (point)))
				 (make-annotation pixmap)))
		    (hypersrc-setup-margin-image anno target-name)
		    (hypersrc-get-new-tag "^IMG")
		    (hypersrc-get-new-tag "^/A$"))))
		 ;; Whether margin comment or margin pixmap, skip to PRE:
		(hypersrc-get-new-tag "^TD") ; Skip "TD"
		(hypersrc-get-new-tag "^PRE$") ; Get "PRE"
		(setq state 'pre)))	; We're back where we started...
	     ((string= "/TABLE" new-tag)
	      (hypersrc-get-new-tag "^/BODY$")
	      (hypersrc-get-new-tag "^/HTML$")
	      (message "Parsing...done.")))))
	 ;; End of the "/PRE" tag case.
	 
	 ((string-match "^A HREF=\"\\([^\"]+\\)\"" new-tag)
	  (progn
	    (setq target (substring new-tag (match-beginning 1) (match-end 1)))
	    (insert (hypersrc-text-between-tags))
	    (hypersrc-get-new-tag "^/A$")
	    (setq ext-beg (point))
	    (insert (hypersrc-text-between-tags))
	    (setq ext-end (point))
	    (setq new-link (make-extent ext-beg ext-end))
	    (set-extent-face new-link 'hypersrc-link-face)
	    (set-extent-property new-link 'keymap hypersrc-link-map)
	    (setq target-file-name
		  (substring target 0 (string-match "#" target )))
	    (if (string-match "#" target )
		(setq target-id-string (substring target (match-end 0)))
	      (setq target-id-string nil))
	    ;; All HyperSource extents are marked with this property:
	    (set-extent-property new-link 'hypersrc 'link)
	    (set-extent-property new-link 'duplicable t)
	    (set-extent-property new-link 'target-file target-file-name)
	    (if target-id-string
		(set-extent-property new-link 'target-id target-id-string))))
	 ((string-match "^A ID=\"\\([^\"]+\\)\"" new-tag)
	  (progn
	    (setq id-string 
		  (substring new-tag (match-beginning 1) (match-end 1)))
	    (insert (hypersrc-text-between-tags))
	    (hypersrc-get-new-tag "^/A$")
	    (setq ext-beg (point))
	    (insert (hypersrc-text-between-tags))
	    (setq ext-end (point))
	    (setq new-target (make-extent ext-beg ext-end))
	    (set-extent-property new-target 'target-id id-string)
	    (set-extent-face new-target 'hypersrc-target-face)
	    (set-extent-property new-target 'duplicable t)))))))
    (hypersrc-replace-escapes)
    (hypersrc-widen-shift-extents)
    ;; The parsed file is now in temp-buf.
    ;; We erase the original buffer, and copy the parsed version back into it.
    ;; After that, we do not need temp-buf anymore, so we can kill it:
    (setq tmp-width (buffer-left-margin-width))
    (set-buffer orig-buf)
    (setq hypersrc-original-left-margin-width (buffer-left-margin-width))
    (set-buffer-left-margin-width tmp-width)
    (setq use-left-overflow t)
    (delete-region (point-min) (point-max))
    (insert-buffer temp-buf)
    (kill-buffer temp-buf)
    (set-buffer-modified-p nil)))
  

(provide 'hypersrc-read)
