;;; hypersrc-write.el -- File saving functions for HyperSource mode

;;; $Id: hypersrc-write.el,v 0.3 1995/05/17 09:17:51 job Exp $

;; Load functions which return the correct comment delimiter depending
;; on the mode passed as argument:
;;
(require 'hypersrc-comments)

(makunbound 'hypersrc-extent-positions-list)
(defvar hypersrc-extent-positions-list ()
  "List of extents that need to be converted into HTML markup.
It contains as elements of the form:
(position is-end extent).
'is-end' is t if this is an end position of an extent, nil otherwise.
Extents are considered only if they have the 'hypersrc' property set.")

(defun hypersrc-record-extent-positions (extent maparg)
  "Add start and end positions of EXTENT to 'hypersrc-extent-positions-list'.
Extents are considered only if they have the 'hypersrc' property set.
Comment and pixmap extents have length 0, so just their start position
is recorded."
  (if (extent-property extent 'hypersrc)
      (progn
	(setq hypersrc-extent-positions-list
	      (cons
	       (list (extent-start-position extent) nil extent)
	       hypersrc-extent-positions-list))
	;; Comment and pixmap extents have a length of 0, so don't
	;; record their end position:
	;; Don't really need this check because map-extents skips
	;; zero-length extents anyway.
	(if (not (or (eq (extent-property extent 'hypersrc) 'comment)
		     (eq (extent-property extent 'hypersrc) 'image)))
	    (setq hypersrc-extent-positions-list
	      (cons
	       (list (extent-end-position extent) t extent)
	       hypersrc-extent-positions-list)))))
  nil)					; Return nil, so map-extents continues

(defun hypersrc-smaller-position (cell1 cell2)
  "Return t if the car of CELL1 is smaller than the car of CELL2."
  (< (car cell1) (car cell2)))

; Unnecessary (annotations have been resized to lenght 1 -> handled w/ extents!)
;(defun hypersrc-build-cell-from-annotation (annotation)
;  "Create (position is-end extent) list from an annotation."
;  (if (extent-property annotation 'hypersrc)
;      (list (extent-start-position annotation) nil annotation)))

(defun hypersrc-sort-extents ()
  "Create sorted list of all positions where an extent starts or ends.
Extents are considered only if they have the 'hypersrc' property set,
meaning that they need to be translated into HTML markup upon saving."
  (setq hypersrc-extent-positions-list ())
  (save-excursion
    (current-buffer)
    (map-extents 'hypersrc-record-extent-positions)
; Unnecessary (annotations have been resized to lenght 1 -> handled w/ extents!)
;    (setq annotations (annotation-list))
;    (setq annotations-cells (mapcar 'hypersrc-build-cell-from-annotation annotations))
;    (setq hypersrc-extent-positions-list (append hypersrc-extent-positions-list annotations-cells))
    (setq hypersrc-extent-positions-list
	  (sort hypersrc-extent-positions-list 'hypersrc-smaller-position))))

(defun hypersrc-next-extent-change (pos &optional buf)
  "Return the next position after POS where an extent begins or ends.\n\
If POS is at the end of the buffer, POS will be returned; otherwise a\n\
position greater than POS will always be returned.\n\
If BUFFER is nil, the current buffer is assumed.\n\
This LISP function should be replaced by a call to the built-in function\n\
'next-extent-change' available in 'extents.c' with XEmacs 19.12."
  (message "NOT IMPLEMENTED"))

;; Variable needed by the following functions:
(setq hypersrc-main-buffer nil)

(defun hypersrc-save-buffer-no-format ()
  "Save HyperSource buffer *without* HTML 3.0 formatting information.
This can be used to create a source file that a compiler understands."
  (interactive)
  ;; Unhook special save function:
  (if (memq 'hypersrc-save-buffer local-write-file-hooks)
      (setq local-write-file-hooks
	    (delq 'hypersrc-save-buffer local-write-file-hooks)))

  ;; Save under file name of the buffer, with final ".hs" removed if existent:
  (write-region
   1
   (1+ (buffer-size))
   (substring buffer-file-name 0 (string-match "\\.hs" buffer-file-name)))

  ;; Hook special save function back in:
  (or (memq 'hypersrc-save-buffer local-write-file-hooks)
      (setq local-write-file-hooks
	    (cons 'hypersrc-save-buffer local-write-file-hooks))))


(defun hypersrc-save-buffer ()
  "Convert a HyperSource buffer into HTML 3.0 format.
This function is hooked into local-write-file-hooks."
  (let (hypersrc-main-buffer hypersrc-current-major-mode
			     hypersrc-main-buffer-size temp-buffer-1 point-1 end-1
			     temp-buffer-2 next-stop is-end next-ext ext-type
			     target-name inline-name)
    (message "Saving as HyperSource...")
    (setq hypersrc-main-buffer (current-buffer)) ; remember name of current buffer
    (setq hypersrc-current-major-mode major-mode)	       ; remember mode of current buffer
    (setq hypersrc-current-left-margin-width (buffer-left-margin-width))
    (setq hypersrc-current-left-margin-pixwidth (buffer-left-margin-pixwidth))
    (save-excursion
      ;; Should we remove possible narrowing here temporarily to really
      ;; get to the buffer's beginning?
      ;; Before release: Start temp buffer names with space to make them
      ;; invisible!
      
      ;; Copy buffer to temp buffer 1. That's where we replace markup chars:
      (setq hypersrc-main-buffer-size (1+ (buffer-size)))
      (setq temp-buffer-1 
	    (generate-new-buffer (concat "*" (buffer-name) "-1*")))
      (set-buffer temp-buffer-1)
      (set-buffer-left-margin-width hypersrc-current-left-margin-width)
      (insert-buffer-substring hypersrc-main-buffer 1 hypersrc-main-buffer-size)
      
      ;; Replace markup characters with escape sequences. '&' must be
      ;; replaced first to avoid double replacements!
      (goto-char 1)
      (while (search-forward "&" (1+ (buffer-size)) t)
	(replace-match "&amp;" nil t))
      (goto-char 1)
      (while (search-forward "<" (1+ (buffer-size)) t)
	(replace-match "&lt;" nil t))
      (goto-char 1)
      (while (search-forward ">" (1+ (buffer-size)) t)
	(replace-match "&gt;" nil t))
      (goto-char 1)
      (setq point-1 (point))
      (setq end-1 (1+ (buffer-size)))
      
      ;; Fill hypersrc-extent-positions-list:
      (hypersrc-sort-extents)
      
      ;; Now build real HTML code in second temp buffer:
      (setq temp-buffer-2
	    (generate-new-buffer (concat "*" (buffer-name hypersrc-main-buffer) "-2*")))
      (set-buffer temp-buffer-2)
      (set-buffer-left-margin-width hypersrc-current-left-margin-width)
      (hypersrc-insert-HTML-doc-start)
      (while (not (eq point-1 end-1))
	(if hypersrc-extent-positions-list ;If list contains another stop point:
	    (progn
	      (setq next-stop (elt (car hypersrc-extent-positions-list) 0)
		    is-end    (elt (car hypersrc-extent-positions-list) 1)
		    next-ext  (elt (car hypersrc-extent-positions-list) 2)
		    hypersrc-extent-positions-list
		    (cdr hypersrc-extent-positions-list))
	      (insert-buffer-substring temp-buffer-1 point-1 next-stop)
	      (setq point-1 next-stop)
	      (setq ext-type (extent-property next-ext 'hypersrc))
	      (if (not is-end)
		  ;; Handle start tags
		  (cond
		   ;; Header
		   ((eq ext-type 'header)
		    (hypersrc-insert-HTML-header-start
		     (extent-property next-ext 'headersize)))
		   ;; Link
		   ((eq ext-type 'link)
		    (hypersrc-insert-HTML-link-start
		     (concat
		      (extent-property next-ext 'target-file)
		      (if (extent-property next-ext 'target-id)
			  (concat "#" (extent-property next-ext 'target-id))))))
		   ;; Link target
		   ((eq ext-type 'target)
		    (hypersrc-insert-HTML-target-start
		     (extent-property next-ext 'target-id)))
		   ;; Comment
		   ((eq ext-type 'comment)
		    (hypersrc-insert-HTML-comment
		     (annotation-glyph next-ext) hypersrc-current-major-mode))
		   ;; Image
		   ((eq ext-type 'image)
		    (progn
		      (setq target-name (cdr (annotation-data next-ext)))
		      (setq
		       inline-name
		       (hypersrc-convert-any-to-gif
			(extent-property next-ext 'target)
			hypersrc-current-left-margin-pixwidth))
		      (hypersrc-insert-HTML-image target-name inline-name))))
		   ;; else: end tag
		   (cond
		    ;; Header
		    ((eq ext-type 'header)
		     (hypersrc-insert-HTML-header-end
		      (extent-property next-ext 'headersize)))
		    ;; Link
		    ((eq ext-type 'link)
		     (hypersrc-insert-HTML-link-end))
		    ;; Target
		    ((eq ext-type 'target)
		     (hypersrc-insert-HTML-target-end)))))
	      ;; else: no more extents ahead...
	      (insert-buffer-substring temp-buffer-1 point-1 end-1)
	      (setq point-1 end-1)
	      ))			; end (while not at end of source buffer)
	(hypersrc-insert-HTML-doc-end)
	(set-visited-file-name (buffer-file-name hypersrc-main-buffer))
	(or buffer-backed-up (backup-buffer))
	(write-region 1 (1+ (buffer-size)) (buffer-file-name hypersrc-main-buffer))
	;; Mark this temp buffer unmodified so it can be killed:
	(set-buffer-modified-p nil))
      
      ;; End of save-excursion
      ;; Before release: Kill temp buffers!
      (kill-buffer temp-buffer-1)
      (kill-buffer temp-buffer-2)
      ;; The main buffer has now been saved, so set it to be unmodified:
      (set-visited-file-modtime)
      (set-buffer-modified-p nil)
      ;; Return non-nil so other save functions are not called:
      t))

    
(defun hypersrc-insert-HTML-doc-start ()
  "Insert HTML code needed at the beginning of a file."
  (insert
   "<!DOCTYPE HTML PUBLIC \"-//IETF//DTD HTML 3.0//EN//\">\n"
   "<!-- HyperSource 1.0 File Format -->\n"
   "<HTML>\n"
   "<HEAD>\n"
   "<TITLE>" (buffer-file-name hypersrc-main-buffer) "</TITLE>\n"
   "</HEAD>\n"
   "\n"
   "<BODY>\n"
   "<TABLE COLSPEC=\"R" (int-to-string hypersrc-current-left-margin-width) " L80\">\n"
   "<TR VALIGN=BASELINE>\n"
   "<TD>\n"
   "<TD CLASS=SOURCE>\n"
   "<PRE>"))

(defun hypersrc-insert-HTML-doc-end ()
  "Insert HTML code needed at the end of a file."
  (insert
   "</PRE>\n"
   "</TABLE>\n"
   "</BODY>\n"
   "</HTML>\n"))

(defun hypersrc-insert-HTML-header-start (size)
  "Insert HTML code for beginning of header.
SIZE (1..6) of header is passed as argument."
  (insert
   "</PRE>\n"
   "<H" (int-to-string size) ">"))

(defun hypersrc-insert-HTML-header-end (size)
  "Insert HTML code for end of header.
SIZE (1..6) of header is passed as argument."
  (insert
   "</H" (int-to-string size) ">\n"
   "<PRE>"))

(defun hypersrc-insert-HTML-link-start (target)
  "Insert HTML code for beginning of hypertext link.
TARGET (in URL notation) is passed as argument."
  (insert
   "<A HREF=\"" target "\">"))

(defun hypersrc-insert-HTML-link-end ()
  "Insert HTML code for end of hypertext link."
  (insert
   "</A>"))

(defun hypersrc-insert-HTML-target-start (id)
  "Insert HTML code for beginning of hypertext link target.
ID is unique identifier passed as argument."
  (insert
   "<A ID=\"" id "\">"))

(defun hypersrc-insert-HTML-target-end ()
  "Insert HTML code for end of hypertext link target."
  (insert
   "</A>"))

(defun hypersrc-insert-HTML-comment (comment-text hypersrc-current-major-mode)
  "Insert HTML code for marginal comment (uses HTML table tags).
COMMENT-TEXT of comment is passed as argument. Characters like '<' are
escaped correctly."
  ;; Replace markup characters with escape sequences. '&' must be
  ;; replaced first to avoid double replacements!
  (setq comment-buffer (generate-new-buffer "*Comment*"))
  (save-excursion
    (set-buffer comment-buffer)
    (insert comment-text)
    (goto-char 1)
    (while (search-forward "&" (1+ (buffer-size)) t)
      (replace-match "&amp;" nil t))
    (goto-char 1)
    (while (search-forward "<" (1+ (buffer-size)) t)
      (replace-match "&lt;" nil t))
    (goto-char 1)
    (while (search-forward ">" (1+ (buffer-size)) t)
      (replace-match "&gt;" nil t))
    (setq comment-text (buffer-substring 1 (1+ (buffer-size)))))
  (kill-buffer comment-buffer)
  ;; Now, insert HTML code & modified text:
  (insert
   "</PRE>\n"
   "<TR VALIGN=BASELINE>\n"
   "<TD CLASS=COMMENT NOWRAP>\n"
   "<SMALL>"
   (hypersrc-open-comment hypersrc-current-major-mode)
   comment-text
   (hypersrc-close-comment hypersrc-current-major-mode)
   "</SMALL>\n"
   "<TD CLASS=SOURCE>\n"
   "<PRE>"))

(defun hypersrc-insert-HTML-image (target-name inline-name)
  "Insert HTML code for marginal image (uses HTML table).
TARGET-NAME (name of original image file) and INLINE-NAME (name of
inline image that serves as preview and link) are passed as arguments."
  (insert
   "</PRE>\n"
   "<TR VALIGN=BASELINE>\n"
   "<TD CLASS=IMAGE><A HREF=\"" target-name "\"><IMG SRC=\"" inline-name "\" ALT=\"\"></A>\n"
   "<TD CLASS=SOURCE>\n"
   "<PRE>"))


(provide 'hypersrc-write)
